// ********************************************************** //
//                                                            //
//          Pascal interface for WinRing0 driver              //
//                  (by OpenLibSys.org)                       //
//                                                            //
//                   Ver.: 2017-12-01                         //
//            Original file name: WinRing0.pas                //
//                                                            //
//    Copyright (c): 2017, NetSpirit (netspirit@meta.ua)      //
//                                                            //
// ********************************************************** //


// ********************** LICENSE *************************** //
//                                                            //
//     Free for use, modify, distribute with preserving       //
//    original copyright info and same license conditions     //
//                                                            //
// ********************************************************** //





unit WinRing0;


{$ifdef CPUX64}
  {$define CPU64}
{$endif}

{$ifndef CPU64}
  {$define CPU32}
{$endif}

interface

uses
  Windows;

const

  OLS_DRIVER_ID = 'WinRing0_1_2_0';


//-----------------------------------------------------------------------------
//
// Version Information
//
//-----------------------------------------------------------------------------
  
  OLS_DRIVER_MAJOR_VERSION  = 1;
  OLS_DRIVER_MINOR_VERSION  = 2;
  OLS_DRIVER_REVISION       = 0;
  OLS_DRIVER_RELESE         = 5;

  OLS_DRIVER_VERSION = ((OLS_DRIVER_MAJOR_VERSION shl 24) or
  (OLS_DRIVER_MINOR_VERSION shl 16) or
  (OLS_DRIVER_REVISION shl 8) or
  OLS_DRIVER_RELESE);


//-----------------------------------------------------------------------------
//
// PCI Error Code
//
//-----------------------------------------------------------------------------

  OLS_ERROR_PCI_BUS_NOT_EXIST = $E0000001;
  OLS_ERROR_PCI_NO_DEVICE     = $E0000002;
  OLS_ERROR_PCI_WRITE_CONFIG  = $E0000003;
  OLS_ERROR_PCI_READ_CONFIG   = $E0000004;



type
  {$ifdef CPU64}
    DWORD_PTR = QWord;
  {$else}
    DWORD_PTR = DWord; 
  {$endif}




 {******************************************************************************
**
** User functions
**
******************************************************************************}
  
  function OpenDriver: Boolean;
  procedure CloseDriver;
  
  
//-----------------------------------------------------------------------------
//
// Support Macros
//
//-----------------------------------------------------------------------------

  
  // Bus Number, Device Number and Function Number to PCI Device Address
  function PCIBusDevFunc(const Bus, Dev, Func: Byte): DWord;
  
  // PCI Device Address to Bus Number
  function PCIGetBus(const Address: DWord): Byte;
  
  // PCI Device Address to Device Number
  function PCIGetDev(const Address: DWord): Byte;
  
  // PCI Device Address to Function Number
  function PCIGetFunc(const Address: DWord): Byte;


 {******************************************************************************
**
** Misc driver functions
**
******************************************************************************}

  function GetDriverVersion: DWord;
  function GetRefCount: DWord;

  
{******************************************************************************
**
** CPU
**
******************************************************************************}
  
  function Rdmsr(
    Index: DWord;          // MSR index
    EAX: PDWord;           // bit  0-31
    EDX: PDWord = nil      // bit 32-63
  ): Boolean;
  
  function RdmsrTx(
    Index: DWord;          // MSR index
    EAX: PDWord;           // bit  0-31
    EDX: PDWord;           // bit 32-63
    ThreadAffinityMask: DWORD_PTR
  ): Boolean;
  
  function RdmsrPx(
    Index: DWord;          // MSR index
    EAX: PDWord;           // bit  0-31
    EDX: PDWord;           // bit 32-63
    ProcessAffinityMask: DWORD_PTR
  ): Boolean;
  
  function Wrmsr(
    Index: DWord;        // MSR index
    EAX: DWord;          // bit  0-31
    EDX: DWord           // bit 32-63
  ): Boolean;
  
  function WrmsrTx(
    Index: DWord;        // MSR index
    EAX: DWord;          // bit  0-31
    EDX: DWord;          // bit 32-63
    ThreadAffinityMask: DWORD_PTR
  ): Boolean;
  
  function WrmsrPx(
    Index: DWord;        // MSR index
    EAX: DWord;          // bit  0-31
    EDX: DWord;          // bit 32-63
    ProcessAffinityMask: DWORD_PTR
  ): Boolean;
  
  function Rdpmc(
    Index: DWord;        // PMC index
    EAX: PDWord;         // bit  0-31
    EDX: PDWord = nil    // bit 32-63
  ): Boolean;
  
  function RdpmcTx(
    Index: DWord;        // PMC index
    EAX: PDWord;         // bit  0-31
    EDX: PDWord;         // bit 32-63
    ThreadAffinityMask: DWORD_PTR
  ): Boolean;
  
  function RdpmcPx(
    Index: DWord;        // PMC index
    EAX: PDWord;         // bit  0-31
    EDX: PDWord;         // bit 32-63
    ProcessAffinityMask: DWORD_PTR
  ): Boolean;
  
  function Hlt(): Boolean;
  
  function HltTx(
    ThreadAffinityMask: DWORD_PTR
  ): Boolean;
  
  function HltPx(
    ProcessAffinityMask: DWORD_PTR
  ): Boolean;


{******************************************************************************
**
** I/O
**
******************************************************************************}

  function ReadIoPortByte(
    Port: Word      // I/O port address
  ): Byte;

  function ReadIoPortWord(
    Port: Word      // I/O port address
  ): Word;

  function ReadIoPortDword(
    Port: Word      // I/O port address
  ): DWord;

  function ReadIoPortByteEx(
    Port: Word;      // I/O port address
    Value: PByte     // Read Value
  ): Boolean;

  function ReadIoPortWordEx(
    Port: Word;      // I/O port address
    Value: PWord     // Read Value
  ): Boolean;

  function ReadIoPortDwordEx(
    Port: Word;      // I/O port address
    Value: PDWord    // Read Value
  ): Boolean;

  procedure WriteIoPortByte(
    Port: Word;      // I/O port address
    Value: Byte      // Write Value
  );

  procedure WriteIoPortDword(
    Port: Word;      // I/O port address
    Value: DWord     // Write Value
  );

  procedure WriteIoPortWord(
    Port: Word;      // I/O port address
    Value: Word      // Write Value
  );

  function WriteIoPortByteEx(
    Port: Word;      // I/O port address
    Value: Byte      // Write Value
  ): Boolean;

  function WriteIoPortWordEx(
    Port: Word;      // I/O port address
    Value: Word      // Write Value
  ): Boolean;

  function WriteIoPortDwordEx(
    Port: Word;      // I/O port address
    Value: DWord     // Write Value
  ): Boolean;


(******************************************************************************
**
** PCI
**
******************************************************************************)
  
  // PCIAddress
  //  0- 2: Function Number
  //  3- 7: Device Number
  //  8-15: PCI Bus Number
  // 16-31: Reserved
  // 0xFFFFFFFF : Error
  
  function ReadPCIConfigByte(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: Byte    // Configuration Address 0-255
  ): Byte;
  
  function ReadPCIConfigWord(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: Byte    // Configuration Address 0-255
  ): Word;
  
  function ReadPCIConfigDword(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: Byte    // Configuration Address 0-255
  ): DWord;
  
  function ReadPCIConfigByteEx(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: DWord;  // Configuration Address 0-whatever
    Value: PByte        // Read Value
  ): Boolean;
  
  function ReadPCIConfigWordEx(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: DWord;  // Configuration Address 0-whatever
    Value: PWord        // Read Value
  ): Boolean;
  
  function ReadPCIConfigDwordEx(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: DWord;  // Configuration Address 0-whatever
    Value: PDWord       // Read Value
  ): Boolean;
  
  procedure WritePCIConfigByte(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: Byte;   // Configuration Address 0-255
    Value: Byte         // Write Value
  );
  
  procedure WritePCIConfigWord(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: Byte;   // Configuration Address 0-255
    Value: Word         // Write Value
  );
  
  procedure WritePCIConfigDword(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: Byte;   // Configuration Address 0-255
    Value: DWord        // Write Value
  );
  
  function WritePCIConfigByteEx(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: DWord;  // Configuration Address 0-whatever
    Value: Byte         // Write Value
  ): Boolean;
  
  function WritePCIConfigWordEx(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: DWord;  // Configuration Address 0-whatever
    Value: Word         // Write Value
  ): Boolean;
  
  function WritePCIConfigDwordEx(
    PCIAddress: DWord;  // PCI Device Address
    RegAddress: DWord;  // Configuration Address 0-whatever
    Value: DWord        // Write Value
  ): Boolean;
  
  // Returns PCIAddress; 0xFFFFFFFF = failure
  function FindPCIDeviceById(
    VendorId: Word;    // Vendor ID
    DeviceId: Word;    // Device ID
    Index: Byte        // Index
  ): DWord;
  
  // Returns PCIAddress; 0xFFFFFFFF = failure
  function FindPCIDeviceByClass(
    BaseClass: Byte;    // Base Class
    SubClass: Byte;     // Sub Class
    ProgramIf: Byte;    // Program Interface
    Index: Byte         // Index
  ): DWord;



{******************************************************************************
**
** Memory (Special API)
**
******************************************************************************}


  // Read size (byte), 0: failure
  function ReadDmiMemory(
    Buffer: Pointer;    // Buffer
    Count: DWord;       // Count
    UnitSize: Byte = 1  // Unit Size (BYTE, WORD, DWORD)
  ): DWord;

  // Read size (byte), 0: failure
  function ReadPhysicalMemory(
    Address: DWORD_PTR;  // Physical Memory Address
    Buffer: Pointer;     // Buffer
    Count: DWord;        // Count
    UnitSize: Byte = 1   // Unit Size (BYTE, WORD, DWORD)
  ): DWord;

  // Write size (byte), 0: failure
  function WritePhysicalMemory(
    Address: DWORD_PTR;   // Physical Memory Address
    Buffer: Pointer;      // Buffer
    Count: DWord;         // Count
    UnitSize: Byte = 1    // Unit Size (BYTE, WORD, DWORD)
  ): DWord;










const
  InvalidPCIAddress = $FFFFFFFF;
  gPCINumberOfDevice =  32;
  gPCINumberOfFunction = 8;
  
var
  gPCIMaxNumberOfBus: Byte = 255; // Max PCI Bus to Scan






implementation






const

//-----------------------------------------------------------------------------
//
// The Device type codes form 32768 to 65535 are for customer use.
//
//-----------------------------------------------------------------------------

  OLS_TYPE = 40000;

//-----------------------------------------------------------------------------
//
// The IOCTL function codes from $800 to $FFF are for customer use.
//
//-----------------------------------------------------------------------------
  
  
  FILE_ANY_ACCESS     = 0;
  FILE_SPECIAL_ACCESS = FILE_ANY_ACCESS;
  FILE_READ_ACCESS    = $0001;           // file & pipe
  FILE_WRITE_ACCESS   = $0002;           // file & pipe
  
  METHOD_BUFFERED   = 0;
  METHOD_IN_DIRECT  = 1;
  METHOD_OUT_DIRECT = 2;
  METHOD_NEITHER    = 3;
  
  
  
  IOCTL_OLS_GET_DRIVER_VERSION =
    (OLS_TYPE shl 16) or(FILE_ANY_ACCESS shl 14) or
    ($800 shl 2) or METHOD_BUFFERED;
 
  IOCTL_OLS_GET_REFCOUNT =
    (OLS_TYPE shl 16) or (FILE_ANY_ACCESS shl 14) or
    ($801 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_MSR =
    (OLS_TYPE shl 16) or (FILE_ANY_ACCESS shl 14) or
    ($821 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_WRITE_MSR =
    (OLS_TYPE shl 16) or (FILE_ANY_ACCESS shl 14) or
    ($822 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_PMC =
    (OLS_TYPE shl 16) or (FILE_ANY_ACCESS shl 14) or
    ($823 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_HALT =
    (OLS_TYPE shl 16) or (FILE_ANY_ACCESS shl 14) or
    ($824 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_IO_PORT =
    (OLS_TYPE shl 16) or (FILE_READ_ACCESS shl 14) or
    ($831 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_WRITE_IO_PORT =
    (OLS_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or
    ($832 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_IO_PORT_BYTE =
    (OLS_TYPE shl 16) or (FILE_READ_ACCESS shl 14) or
    ($833 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_IO_PORT_WORD =
    (OLS_TYPE shl 16) or (FILE_READ_ACCESS shl 14) or
    ($834 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_IO_PORT_DWORD =
    (OLS_TYPE shl 16) or (FILE_READ_ACCESS shl 14) or
    ($835 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_WRITE_IO_PORT_BYTE =
    (OLS_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or
    ($836 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_WRITE_IO_PORT_WORD =
    (OLS_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or
    ($837 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_WRITE_IO_PORT_DWORD =
    (OLS_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or
    ($838 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_MEMORY =
    (OLS_TYPE shl 16) or (FILE_READ_ACCESS shl 14) or
    ($841 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_WRITE_MEMORY =
    (OLS_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or
    ($842 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_READ_PCI_CONFIG =
    (OLS_TYPE shl 16) or (FILE_READ_ACCESS shl 14) or
    ($851 shl 2) or METHOD_BUFFERED;

  IOCTL_OLS_WRITE_PCI_CONFIG =
    (OLS_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or
    ($852 shl 2) or METHOD_BUFFERED;




type
  
  OLS_WRITE_MSR_INPUT = packed record
    Reg: DWord;
    Value: LARGE_INTEGER;
  end;

  OLS_WRITE_IO_PORT_INPUT = packed record
    PortNumber: DWord; 
    case Integer of
      0: (LongData: DWord);
      1: (ShortData: Word);
      2: (ByteData: Byte);
  end;

  OLS_READ_PCI_CONFIG_INPUT = packed record
    PCIAddress: DWord;
    PCIOffset: DWord;
  end;

  {OLS_WRITE_PCI_CONFIG_INPUT = record
    PCIAddress: DWord;
    PCIOffset: DWord;
    Data: array[0..0] of Byte; // Record with variable data length at the end (not used)
  end;}
  
  OLS_WRITE_PCI_CONFIG_INPUT = packed record
    PCIAddress: DWord;
    PCIOffset: DWord;
    Data: DWord;
  end;

  PHYSICAL_ADDRESS = LARGE_INTEGER;

  OLS_READ_MEMORY_INPUT = packed record
    Address: PHYSICAL_ADDRESS;
    UnitSize: DWord;
    Count: DWord;
  end;

  OLS_WRITE_MEMORY_INPUT = packed record
    Address: PHYSICAL_ADDRESS;   
    UnitSize: DWord;
    Count: DWord;
    Data: array [0..0] of Byte;
  end;






var
  hDriver: THandle;
  RefCount: DWord;




function OpenDriver: Boolean;
begin
  Result := True;
  if RefCount = 0 then
  begin
    hDriver := CreateFile(
      PChar('\\.\' + OLS_DRIVER_ID),
      GENERIC_READ or GENERIC_WRITE,
      0,
      nil,
      OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL,
      0
    );
    if hDriver = INVALID_HANDLE_VALUE then
      Result := False
    else
      Inc(RefCount);
  end
  else
    Inc(RefCount);
end;

procedure CloseDriver;
begin
  if RefCount = 0 then Exit;

  Dec(RefCount);
  
  if RefCount = 0 then
  begin
    CloseHandle(hDriver);
    hDriver := INVALID_HANDLE_VALUE;
  end;
end;





function PCIBusDevFunc(const Bus, Dev, Func: Byte): DWord; {$ifndef VER120} inline;{$endif}
begin
  Result := (Bus {and $FF)} shl 8) or ((Dev and $1F) shl 3) or (Func and 7);
end;

function PCIGetBus(const Address: DWord): Byte;{$ifndef VER120} inline;{$endif}
begin
  Result := ((Address shr 8) {and $FF});
end;

function PCIGetDev(const Address: DWord): Byte;{$ifndef VER120} inline;{$endif}
begin
  Result := ((Address shr 3) {and $1F});
end;

function PCIGetFunc(const Address: DWord): Byte;{$ifndef VER120} inline;{$endif}
begin
  Result := (Address and 7);
end;



//-----------------------------------------------------------------------------
//
// Misc driver functions
//
//-----------------------------------------------------------------------------


function GetDriverVersion: DWord;
var
  L: DWord;
begin
  Result := 0;
  
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_GET_DRIVER_VERSION),
    nil,
    0,
    @Result,
    SizeOf(Result),
    L,
    nil
  );
end;

function GetRefCount: DWord;
var
  L: DWord;
begin
  Result := 0;
  
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_GET_REFCOUNT),
    nil,
    0,
    @Result,
    SizeOf(Result),
    L,
    nil
  );
end;


//-----------------------------------------------------------------------------
//
// CPU
//
//-----------------------------------------------------------------------------


function Rdmsr(Index: DWord; EAX: PDWord; EDX: PDWord = nil): Boolean;
var
  L: DWord;
  //OutBuf: array [0..7] of Byte;
  OutBuf: array [0..1] of DWord;
begin
  //Result := False;
  
  if (hDriver = INVALID_HANDLE_VALUE) or
     (EAX = nil)
  then
  begin
    Result := False;
    Exit;
  end;
  
  EAX^ := 0;
  if EDX <> nil then EDX^ := 0;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_MSR),
    @index,
    SizeOf(index),
    @OutBuf,
    SizeOf(OutBuf),
    L,
    nil
  );
  
  if Result then
  begin
    //Move(OutBuf[0], EAX^, 4);
    //if EDX <> nil then Move(OutBuf[4], EDX^, 4);
    EAX^ := OutBuf[0];
    if EDX <> nil then EDX^ := OutBuf[1];
  end;
end;

function RdmsrTx(Index: DWord; EAX: PDWord; EDX: PDWord; ThreadAffinityMask: DWORD_PTR): Boolean;
var
  hThread: THandle;
  Mask: DWORD_PTR;
begin
  hThread := GetCurrentThread();
  Mask := SetThreadAffinityMask(hThread, ThreadAffinityMask);
  Result := Rdmsr(Index, EAX, EDX);
  SetThreadAffinityMask(hThread, Mask);
end;

function RdmsrPx(Index: DWord; EAX: PDWord; EDX: PDWord; ProcessAffinityMask: DWORD_PTR): Boolean;
var
  hProcess: THandle;
  ProcessMask: DWORD_PTR;
  SystemMask: DWORD_PTR;
begin
  hProcess := GetCurrentProcess();
  GetProcessAffinityMask(hProcess, ProcessMask, SystemMask);
  SetProcessAffinityMask(hProcess, ProcessAffinityMask);
  Result := Rdmsr(Index, EAX, EDX);
  SetProcessAffinityMask(hProcess, ProcessMask);
end;

function Wrmsr(Index: DWord; EAX: DWord; EDX: DWord): Boolean;
var
  L: DWord;
  OutBuf: DWord;
  InBuf: OLS_WRITE_MSR_INPUT;
begin
  
  if hDriver = INVALID_HANDLE_VALUE then
  begin
    Result := False;
    Exit;
  end;
  
  InBuf.Reg := Index;
  InBuf.Value.HighPart := EDX;
  InBuf.Value.LowPart := EAX;
  
  L := 0;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_MSR),
    @InBuf,
    SizeOf(InBuf),
    @OutBuf,
    SizeOf(OutBuf),
    L,
    nil
  );
end;

function WrmsrTx(Index: DWord; EAX: DWord; EDX: DWord; ThreadAffinityMask: DWORD_PTR): Boolean;
var
  hThread: THandle;
  Mask: DWORD_PTR;
begin
  hThread := GetCurrentThread();
  Mask := SetThreadAffinityMask(hThread, ThreadAffinityMask);
  Result := Wrmsr(Index, EAX, EDX);
  SetThreadAffinityMask(hThread, Mask);
end;

function WrmsrPx(Index: DWord; EAX: DWord; EDX: DWord; ProcessAffinityMask: DWORD_PTR): Boolean;
var
  hProcess: THandle;
  ProcessMask: DWORD_PTR;
  SystemMask: DWORD_PTR;
begin
  hProcess := GetCurrentProcess();
  GetProcessAffinityMask(hProcess, ProcessMask, SystemMask);
  SetProcessAffinityMask(hProcess, ProcessAffinityMask);
  Result := Wrmsr(Index, EAX, EDX);
  SetProcessAffinityMask(hProcess, ProcessMask);
end;

function Rdpmc(Index: DWord; EAX: PDWord; EDX: PDWord = nil): Boolean;
var
  L: DWord;
  //OutBuf: array [0..7] of Byte;
  OutBuf: array [0..1] of DWord;
begin
  //Result := False;
  
  if (hDriver = INVALID_HANDLE_VALUE) or
     (EAX = nil)
  then
  begin
    Result := False;
    Exit;
  end;
  
  EAX^ := 0;
  if EDX <> nil then EDX^ := 0;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_PMC),
    @Index,
    SizeOf(Index),
    @OutBuf,
    SizeOf(OutBuf),
    L,
    nil
  );
  
  if Result then
  begin
    //Move(OutBuf[0], EAX^, 4);
    //Move(OutBuf[4], EDX^, 4);
    EAX^ := OutBuf[0];
    if EDX <> nil then EDX^ := OutBuf[1];
  end;
end;

function RdpmcTx(Index: DWord; EAX: PDWord; EDX: PDWord; ThreadAffinityMask: DWORD_PTR): Boolean;
var
  hThread: THandle;
  Mask: DWORD_PTR;
begin
  hThread := GetCurrentThread();
  Mask := SetThreadAffinityMask(hThread, ThreadAffinityMask);
  Result := Rdpmc(Index, EAX, EDX);
  SetThreadAffinityMask(hThread, Mask);
end;

function RdpmcPx(Index: DWord; EAX: PDWord; EDX: PDWord; ProcessAffinityMask: DWORD_PTR): Boolean;
var
  hProcess: THandle;
  ProcessMask: DWORD_PTR;
  SystemMask: DWORD_PTR;
begin
  hProcess := GetCurrentProcess();
  GetProcessAffinityMask(hProcess, ProcessMask, SystemMask);
  SetProcessAffinityMask(hProcess, ProcessAffinityMask);
  Result := Rdpmc(Index, EAX, EDX);
  SetProcessAffinityMask(hProcess, ProcessMask);
end;

function Hlt(): Boolean;
var
  L: DWord;
begin
  if hDriver = INVALID_HANDLE_VALUE then
  begin
    Result := False;
    Exit;
  end;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_HALT),
    nil,
    0,
    nil,
    0,
    L,
    nil
  );
end;

function HltTx(ThreadAffinityMask: DWORD_PTR): Boolean;
var
  hThread: THandle;
  Mask: DWORD_PTR;
begin
  hThread := GetCurrentThread();
  Mask := SetThreadAffinityMask(hThread, ThreadAffinityMask);
  Result := Hlt();
  SetThreadAffinityMask(hThread, Mask);
end;

function HltPx(ProcessAffinityMask: DWORD_PTR): Boolean;
var
  hProcess: THandle;
  ProcessMask: DWORD_PTR;
  SystemMask: DWORD_PTR;
begin
  hProcess := GetCurrentProcess();
  GetProcessAffinityMask(hProcess, ProcessMask, SystemMask);
  SetProcessAffinityMask(hProcess, ProcessAffinityMask);
  Result := Hlt();
  SetProcessAffinityMask(hProcess, ProcessMask);
end;


//-----------------------------------------------------------------------------
// I/O Read (No Error Check)
//-----------------------------------------------------------------------------

function ReadIoPortByte(Port: Word): Byte;
var
  L: DWord;
begin
  Result := 0;
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_IO_PORT_BYTE),
    @Port,
    SizeOf(Port),
    @Result,
    SizeOf(Result),
    L,
    nil
  );
end;

function ReadIoPortWord(Port: Word): Word;
var
  L: DWord;
begin
  Result := 0;
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_IO_PORT_WORD),
    @Port,
    SizeOf(Port),
    @Result,
    SizeOf(Result),
    L,
    nil
  );
end;

function ReadIoPortDword(Port: Word): DWord;
var
  L: DWord;
begin
  Result := 0;
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_IO_PORT_WORD),
    @Port,
    SizeOf(Port),
    @Result,
    SizeOf(Result),
    L,
    nil
  );
end;

//-----------------------------------------------------------------------------
// I/O Read (Error Check)
//-----------------------------------------------------------------------------

function ReadIoPortByteEx(Port: Word; Value: PByte): Boolean;
var
  L: DWord;
begin
  if (hDriver = INVALID_HANDLE_VALUE) or (Value = nil) then
  begin
    Result := False;
    Exit;
  end;
  
  Value^ := 0;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_IO_PORT_BYTE),
    @Port,
    SizeOf(Port),
    @Result,
    SizeOf(Result),
    L,
    nil
  );
end;

function ReadIoPortWordEx(Port: Word; Value: PWord): Boolean;
var
  L: DWord;
begin
  if (hDriver = INVALID_HANDLE_VALUE) or (Value = nil) then
  begin
    Result := False;
    Exit;
  end;
  
  Value^ := 0;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_IO_PORT_WORD),
    @Port,
    SizeOf(Port),
    Value,
    SizeOf(Value^),
    L,
    nil
  );
end;

function ReadIoPortDwordEx(Port: Word; Value: PDWord): Boolean;
var
  L: DWord;
begin
  if (hDriver = INVALID_HANDLE_VALUE) or (Value = nil) then
  begin
    Result := False;
    Exit;
  end;
  
  Value^ := 0;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_IO_PORT_DWORD),
    @Port,
    SizeOf(Port),
    Value,
    SizeOf(Value^),
    L,
    nil
  );
end;

//-----------------------------------------------------------------------------
// I/O Write (No Error Check)
//-----------------------------------------------------------------------------

procedure WriteIoPortByte(Port: Word; Value: Byte);
var
  L, InLen: DWord;
  InBuf: OLS_WRITE_IO_PORT_INPUT;
begin
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  InLen := SizeOf(InBuf) - SizeOf(InBuf.LongData) + SizeOf(Value);
  InBuf.ByteData := Value;
  InBuf.PortNumber := Port;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_IO_PORT_BYTE),
    @InBuf,
    InLen,
    nil,
    0,
    L,
    nil
  );
end;

procedure WriteIoPortWord(Port: Word; Value: Word);
var
  L, InLen: DWord;
  InBuf: OLS_WRITE_IO_PORT_INPUT;
begin
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  InLen := SizeOf(InBuf) - SizeOf(InBuf.LongData) + SizeOf(Value);
  InBuf.ShortData := Value;
  InBuf.PortNumber := Port;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_IO_PORT_WORD),
    @InBuf,
    InLen,
    nil,
    0,
    L,
    nil
  );
end;

procedure WriteIoPortDword(Port: Word; Value: DWord);
var
  L, InLen: DWord;
  InBuf: OLS_WRITE_IO_PORT_INPUT;
begin
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  InLen := SizeOf(InBuf) {- SizeOf(InBuf.LongData) + SizeOf(Value)};
  InBuf.LongData := Value;
  InBuf.PortNumber := Port;
  
  DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_IO_PORT_DWORD),
    @InBuf,
    InLen,
    nil,
    0,
    L,
    nil
  );
end;


//-----------------------------------------------------------------------------
// I/O Write (Error Check)
//-----------------------------------------------------------------------------

function WriteIoPortByteEx(Port: Word; Value: Byte): Boolean;
var
  L, InLen: DWord;
  InBuf: OLS_WRITE_IO_PORT_INPUT;
begin
  if hDriver = INVALID_HANDLE_VALUE then
  begin
    Result := False;
    Exit;
  end;
  
  InLen := SizeOf(InBuf) - SizeOf(InBuf.LongData) + SizeOf(Value);
  InBuf.ByteData := Value;
  InBuf.PortNumber := Port;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_IO_PORT_BYTE),
    @InBuf,
    InLen,
    nil,
    0,
    L,
    nil
  );
end;

function WriteIoPortWordEx(Port: Word; Value: Word): Boolean;
var
  L, InLen: DWord;
  InBuf: OLS_WRITE_IO_PORT_INPUT;
begin
  if hDriver = INVALID_HANDLE_VALUE then
  begin
    Result := False;
    Exit;
  end;
  
  InLen := SizeOf(InBuf) - SizeOf(InBuf.LongData) + SizeOf(Value);
  InBuf.ShortData := Value;
  InBuf.PortNumber := Port;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_IO_PORT_WORD),
    @InBuf,
    InLen,
    nil,
    0,
    L,
    nil
  );
end;

function WriteIoPortDwordEx(Port: Word; Value: DWord): Boolean;
var
  L, InLen: DWord;
  InBuf: OLS_WRITE_IO_PORT_INPUT;
begin
  if hDriver = INVALID_HANDLE_VALUE then
  begin
    Result := False;
    Exit;
  end;
  
  InLen := SizeOf(InBuf) {- SizeOf(InBuf.LongData) + SizeOf(Value)};
  InBuf.LongData := Value;
  InBuf.PortNumber := Port;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_IO_PORT_DWORD),
    @InBuf,
    InLen,
    nil,
    0,
    L,
    nil
  );
end;


//-----------------------------------------------------------------------------
//
// PCI Configuration Read
//
//-----------------------------------------------------------------------------

function PCIConfigRead(PCIAddress: DWord; RegAddress: DWord; Value: Pointer; Size: DWord; Error: PDWord): Boolean;
var
  L: DWord;
  InBuf: OLS_READ_PCI_CONFIG_INPUT;
begin
  Result := False;
  
  if (hDriver = INVALID_HANDLE_VALUE) or (Value = nil) then Exit;
  
  // Alignment check
  if (Size = 2) and ((RegAddress and 1) <> 0) then Exit;
  if (Size = 4) and ((RegAddress and 3) <> 0) then Exit;
  
  InBuf.PCIAddress := PCIAddress;
  InBuf.PCIOffset := RegAddress;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_READ_PCI_CONFIG),
    @InBuf,
    SizeOf(InBuf),
    Value,
    Size,
    L,
    nil
  );
  
  if (Error <> nil) then Error^ := GetLastError();
end;

function ReadPCIConfigByte(PCIAddress: DWord; RegAddress: Byte): Byte;
begin
  Result := 0;
  
  if not (PCIConfigRead(PCIAddress, RegAddress, @Result, SizeOf(Result), nil))
  then
    Result := $FF;
end;

function ReadPCIConfigWord(PCIAddress: DWord; RegAddress: Byte): Word;
begin
  Result := 0;
  
  if not (PCIConfigRead(PCIAddress, RegAddress, @Result, SizeOf(Result), nil))
  then
    Result := $FFFF;
end;

function ReadPCIConfigDword(PCIAddress: DWord; RegAddress: Byte): DWord;
begin
  Result := 0;
  
  if not (PCIConfigRead(PCIAddress, RegAddress, @Result, SizeOf(Result), nil))
  then
    Result := $FFFFFFFF;
end;

function ReadPCIConfigByteEx(PCIAddress: DWord; RegAddress: DWord; Value: PByte): Boolean; {$ifndef VER120} inline;{$endif}
begin
  Result := PCIConfigRead(PCIAddress, RegAddress, Value, SizeOf(Byte), nil);
end;

function ReadPCIConfigWordEx(PCIAddress: DWord; RegAddress: DWord; Value: PWord): Boolean; {$ifndef VER120} inline;{$endif}
begin
  Result := PCIConfigRead(PCIAddress, RegAddress, Value, SizeOf(Word), nil);
end;

function ReadPCIConfigDwordEx(PCIAddress: DWord; RegAddress: DWord; Value: PDWord): Boolean; {$ifndef VER120} inline;{$endif}
begin
  Result := PCIConfigRead(PCIAddress, RegAddress, Value, SizeOf(DWord), nil);
end;


//-----------------------------------------------------------------------------
//
// PCI Configuration Write
//
//-----------------------------------------------------------------------------

function PCIConfigWrite(PCIAddress: DWord; RegAddress: DWord; Value: DWord; Size: Byte): Boolean;
var
  L, InLen: DWord;
  //InBuf: ^OLS_WRITE_PCI_CONFIG_INPUT;
  InBuf: OLS_WRITE_PCI_CONFIG_INPUT;
begin
  Result := False;
  
  if (hDriver = INVALID_HANDLE_VALUE) or
    {(Value = nil) or} (Size = 0) or (Size > 4)
  then
    Exit;
  
  // alignment check
  if (Size = 2) and ((RegAddress and 1) <> 0) then Exit;
  if (Size = 4) and ((RegAddress and 3) <> 0) then Exit;
  
  // Record with variable data length at the end (not used)
  {InLen := SizeOf(InBuf^) - SizeOf(InBuf^.Data) + Size; // (InBuf^.Data: array[0..0] of Byte) will be equal to real size
  GetMem(InBuf, InLen);
  Move(Value^, InBuf^.Data, Size);
  .....
  FreeMem(InBuf);}
  
  InLen := SizeOf(InBuf) - SizeOf(InBuf.Data) + Size;
  
  InBuf.Data := Value;
  InBuf.PCIAddress := PCIAddress;
  InBuf.PCIOffset := RegAddress;
  
  Result := DeviceIoControl(
    hDriver,
    DWord(IOCTL_OLS_WRITE_PCI_CONFIG),
    @InBuf,
    InLen,
    nil,
    0,
    L,
    nil
  );
end;


//-----------------------------------------------------------------------------
// PCI Configuration Write (No Error Check)
//-----------------------------------------------------------------------------

procedure WritePCIConfigByte(PCIAddress: DWord; RegAddress: Byte; Value: Byte); {$ifndef VER120} inline;{$endif}
begin
  PCIConfigWrite(PCIAddress, RegAddress, Value, SizeOf(Value));
end;

procedure WritePCIConfigWord(PCIAddress: DWord; RegAddress: Byte; Value: Word); {$ifndef VER120} inline;{$endif}
begin
  PCIConfigWrite(PCIAddress, RegAddress, Value, SizeOf(Value));
end;

procedure WritePCIConfigDword(PCIAddress: DWord; RegAddress: Byte; Value: DWord); {$ifndef VER120} inline;{$endif}
begin
  PCIConfigWrite(PCIAddress, RegAddress, Value, SizeOf(Value));
end;

//-----------------------------------------------------------------------------
// PCI Configuration Write (Error Check)
//-----------------------------------------------------------------------------

function WritePCIConfigByteEx(PCIAddress: DWord; RegAddress: DWord; Value: Byte): Boolean; {$ifndef VER120} inline;{$endif}
begin
  Result := PCIConfigWrite(PCIAddress, RegAddress, Value, SizeOf(Value));
end;

function WritePCIConfigWordEx(PCIAddress: DWord; RegAddress: DWord; Value: Word): Boolean; {$ifndef VER120} inline;{$endif}
begin
  Result := PCIConfigWrite(PCIAddress, RegAddress, Value, SizeOf(Value));
end;

function WritePCIConfigDwordEx(PCIAddress: DWord; RegAddress: DWord; Value: DWord): Boolean; {$ifndef VER120} inline;{$endif}
begin
  Result := PCIConfigWrite(PCIAddress, RegAddress, Value, SizeOf(Value));
end;


//-----------------------------------------------------------------------------
//
// FindPCIDevice
//
//-----------------------------------------------------------------------------

function FindPCIDeviceById(VendorId: Word; DeviceId: Word; Index: Byte): DWord;
var
  Bus, Dev, Func: Byte;
  Count: Byte;
  PCIAddress: DWord;
  Id: DWord;
  MultiFuncFlag: Boolean;
  Tp: Byte;
begin
  Result := InvalidPCIAddress;
  
  if (hDriver = INVALID_HANDLE_VALUE) or (VendorId = $FFFF) then Exit;
  
  Count := 0;
  
  for Bus := 0 to gPCIMaxNumberOfBus do
  begin
    for Dev := 0 to gPCINumberOfDevice - 1 do
    begin
      MultiFuncFlag := False;
      for Func := 0 to gPCINumberOfFunction - 1 do
      begin
        if (MultiFuncFlag = False) and (Func > 0) then Break;

        PCIAddress := PCIBusDevFunc(Bus, Dev, Func);
        
        if PCIConfigRead(PCIAddress, 0, @Id, SizeOf(Id), nil) then
        begin
          if Func = 0 then // Is Multi Function Device
          begin
            if PCIConfigRead(PCIAddress, $0E, @Tp, SizeOf(Tp), nil) then
            begin
              if (Tp and $80) <> 0 then
                MultiFuncFlag := True;
            end;
          end;

          if (Id = (VendorId or (DeviceId shl 16))) then
          begin
            if Count = Index then
            begin
              Result := PCIAddress;
              Exit;
            end;
            Inc(Count);
            Continue;
          end;
        end;
        
      end;
    end;
  end;
  
end;


function FindPCIDeviceByClass(BaseClass: Byte; SubClass: Byte; ProgramIf: Byte; Index: Byte): DWord;
var
  Bus, Dev, Func: Byte;
  Count: Byte;
  Cls: DWord;
  PCIAddress: DWord;
  Conf: array [0..2] of DWord;
  MultiFuncFlag: Boolean;
  Tp: Byte;
begin
  Result := InvalidPCIAddress;
  
  if hDriver = INVALID_HANDLE_VALUE then Exit;
  
  Count := 0;
  Cls := (BaseClass shl 24) or
         (SubClass shl 16) or
         (ProgramIf shl 8);
  
  for Bus := 0 to gPCIMaxNumberOfBus do
  begin
    for Dev := 0 to gPCINumberOfDevice - 1 do
    begin
      MultiFuncFlag := False;
      for Func := 0 to gPCINumberOfFunction - 1 do
      begin
        if (MultiFuncFlag = False) and (Func > 0) then Break;

        PCIAddress := PCIBusDevFunc(Bus, Dev, Func);
        
        if PCIConfigRead(PCIAddress, 0, @Conf, SizeOf(Conf), nil) then
        begin
          if Func = 0 then // Is Multi Function Device
          begin
            if PCIConfigRead(PCIAddress, $0E, @Tp, SizeOf(Tp), nil) then
            begin
              if (Tp and $80) <> 0 then
                MultiFuncFlag := True;
            end;
          end;

          if (Conf[2] and $FFFFFF00) = Cls then
          begin
            if Count = Index then
            begin
              Result := PCIAddress;
              Exit;
            end;
            Inc(Count);
            Continue;
          end;
        end;
        
      end;
    end;
  end;
end;


//-----------------------------------------------------------------------------
//
// Physical Memory
//
//-----------------------------------------------------------------------------


function ReadDmiMemory(Buffer: Pointer; Count: DWord; UnitSize: Byte = 1): DWord;
var
  L, Size: DWord;
  InBuf: OLS_READ_MEMORY_INPUT;
begin
  Result := 0;
  
  if (hDriver = INVALID_HANDLE_VALUE) or
    (Buffer = nil) or (Count = 0)
  then
    Exit;
  
  {if(SizeOf(DWORD_PTR) = 4) then
  begin
    InBuf.Address.HighPart := 0;
    InBuf.Address.LowPart := (DWord)$000F0000;
  end
  else
  begin
    InBuf.Address.QuadPart = $000F0000;
  end;}
  
  {$ifdef CPU64}
    InBuf.Address.QuadPart := $000F0000;
  {$else}
    InBuf.Address.HighPart := 0;
    InBuf.Address.LowPart := $000F0000;  
  {$endif}
  
  InBuf.UnitSize := UnitSize;
  InBuf.Count := Count;
  Size := InBuf.UnitSize * InBuf.Count;
  
  if Size > 65536 then Exit;
  
  if DeviceIoControl(
      hDriver,
      DWord(IOCTL_OLS_READ_MEMORY),
      @InBuf,
      SizeOf(InBuf),
      Buffer,
      Size,
      L,
      nil
    )
    and (L = Size)
  then
    Result := Size;
end;

function ReadPhysicalMemory(Address: DWORD_PTR; Buffer: Pointer; Count: DWord; UnitSize: Byte = 1): DWord;
var
  L, Size: DWord;
  InBuf: OLS_READ_MEMORY_INPUT;
begin
  Result := 0;
  
  if (hDriver = INVALID_HANDLE_VALUE) or
    (Buffer = nil) or (Count = 0)
  then
    Exit;
  
  {$ifdef CPU64}
    InBuf.Address.QuadPart := Address;
  {$else}
    InBuf.Address.HighPart := 0;
    InBuf.Address.LowPart := Address;  
  {$endif}
  
  InBuf.UnitSize := UnitSize;
  InBuf.Count := Count;
  Size := InBuf.UnitSize * InBuf.Count;
  
  if Size > 65536 then Exit;
  
  if DeviceIoControl(
      hDriver,
      DWord(IOCTL_OLS_READ_MEMORY),
      @InBuf,
      SizeOf(InBuf),
      Buffer,
      Size,
      L,
      nil
    )
    and (L = Size)
  then
    Result := Size;
end;

function WritePhysicalMemory(Address: DWORD_PTR; Buffer: Pointer; Count: DWord; UnitSize: Byte = 1): DWord;
var
  L, Size, InLen: DWord;
  InBuf: ^OLS_WRITE_MEMORY_INPUT;
begin
  Result := 0;
  
  if (hDriver = INVALID_HANDLE_VALUE) or
    (Buffer = nil) or (Count = 0)
  then
    Exit;
  
  Size := UnitSize * Count;
  InLen := SizeOf(OLS_WRITE_MEMORY_INPUT) - 1 + Size; // 1 = SizeOf(OLS_WRITE_MEMORY_INPUT.Data: array [0..0] of Byte)
  
  GetMem(InBuf, InLen);
  
  InBuf^.UnitSize := UnitSize;
  InBuf^.Count := Count;
  
  {$ifdef CPU64}
    InBuf^.Address.QuadPart := Address;
  {$else}
    InBuf^.Address.HighPart := 0;
    InBuf^.Address.LowPart := Address;  
  {$endif}
  
  
  Move(Buffer^, InBuf^.Data, Size);
  
  if DeviceIoControl(
      hDriver,
      DWord(IOCTL_OLS_WRITE_MEMORY),
      @InBuf,
      InLen,
      nil,
      0,
      L,
      nil
    )
  then
    Result := Size;
  
  FreeMem(InBuf);
end;


end.

