unit hiCPUTemperature;


{$ifndef CPU64} {$define CPU32} {$endif}


interface

uses
 Windows, KOL, Share, Debug, WinRing0;
 
const
  // Intel
  // MSR's:
  IA32_THERM_STATUS = $19C;
  MSR_TEMPERATURE_TARGET = $1A2;
  IA32_PACKAGE_THERM_STATUS = $1B1;

  
  // AMD
  PCI_BUS = 0;
  PCI_BASE_DEVICE = $18;
  DEVICE_VENDOR_ID_REGISTER = 0;
  AMD_VENDOR_ID = $1022;
  
  MISCELLANEOUS_CONTROL_FUNCTION = 3;
  MISCELLANEOUS_CONTROL_DEVICE_ID = $1103;
  FAMILY_10H_MISCELLANEOUS_CONTROL_DEVICE_ID = $1203;
  FAMILY_11H_MISCELLANEOUS_CONTROL_DEVICE_ID = $1303;
  FAMILY_12H_MISCELLANEOUS_CONTROL_DEVICE_ID = $1703;
  FAMILY_14H_MISCELLANEOUS_CONTROL_DEVICE_ID = $1703;
  FAMILY_15H_MODEL_00_MISC_CONTROL_DEVICE_ID = $1603;
  FAMILY_15H_MODEL_10_MISC_CONTROL_DEVICE_ID = $1403;
  FAMILY_16H_MODEL_00_MISC_CONTROL_DEVICE_ID = $1533;
  // PCI register
  THERMTRIP_STATUS_REGISTER = $E4;
  REPORTED_TEMPERATURE_CONTROL_REGISTER = $A4;
  

  
  // CPUID
  CPU_TYPE_INTEL     = 1;
  CPU_TYPE_AMD       = 2;
  
  VendorIDIntel: array [0..11] of AnsiChar = 'GenuineIntel';
  VendorIDAMD: array [0..11] of AnsiChar = 'AuthenticAMD';
  ManufacturerINTEL = 'Intel';
  ManufacturerAMD = 'AMD';
  
  BIT_28      = $10000000;
  INTEL_HTT    = BIT_28; // Hyper-threading technology
  AMD_HTT     = BIT_28; // Hyper-Threading Technology
  
  
type

  TIntelSpecific = record
    BrandID: Byte;
    APICID: Byte;
    ExFeatures: Cardinal;
    Ex64Features: Cardinal;
    Ex64Features2: Cardinal;
  end;

  TAMDSpecific = packed record
    ExFeatures: Cardinal;
    ExFeatures2: Cardinal;
    Features2: Cardinal;
    BrandID: Byte;
    APICID: Byte;
    ExBrandID: Word;
  end;

  TCpuInfo = record
    HasInstruction: Boolean;
    HasExtendedInfo: Boolean;
    PType: Byte;
    BaseFamily: Byte;
    ExtendedFamily: Byte;
    Family: Byte;
    BaseModel: Byte;
    ExtendedModel: Byte;
    Model: Byte;
    Stepping: Byte;
    Features: Cardinal;
    VendorIDString: array [0..11] of AnsiChar;
    Manufacturer: array [0..9] of AnsiChar;
    CpuName: array [0..47] of AnsiChar;
    LogicalCore: Byte;
    PhysicalCore: Byte;
    HyperThreadingTechnology: Boolean;
    HardwareHyperThreadingTechnology: Boolean;
    case CpuType: Byte of
      CPU_TYPE_INTEL: (IntelSpecific: TIntelSpecific;);
      CPU_TYPE_AMD: (AMDSpecific: TAMDSpecific;);
  end;


  TCoreTemps = array of Integer; 

  ThiCPUTemperature = class(TDebug)
    private
      FDrvStatus: Integer;
      FTemp: Integer;
    public
      _event_onGetPackageTemp: THI_Event;
      _event_onGetCoreTemp: THI_Event;
      _event_onErrorDriver: THI_Event;
      
      constructor Create;
      destructor Destroy; override;

      procedure _work_doGetPackageTemp(var _Data: TData; Index: Word);
      procedure _work_doGetCoreTemp(var _Data: TData; Index: Word);
      
      procedure _var_CPUTemp(var _Data: TData; Index: Word);
      procedure _var_VendorID(var _Data: TData; Index: Word);
      procedure _var_CPUName(var _Data: TData; Index: Word);
      procedure _var_Family(var _Data: TData; Index: Word);
      procedure _var_Model(var _Data: TData; Index: Word);
      procedure _var_Stepping(var _Data: TData; Index: Word);
      procedure _var_PhysicalCores(var _Data: TData; Index: Word);
      procedure _var_LogicalCores(var _Data: TData; Index: Word);
  end;

implementation

type
  TPackageTempProc = function(): Integer;
  TCoreTempProc = function(): TCoreTemps;

var
  CpuInfo: TCpuInfo;
  
  // Temperature params
  IntelTJMax: Integer; // = 100;
  IntelTSlope: Integer = 1;
  
  AMDtOffset: Integer = -49;
  AMDPciAddress: DWord;
  AMDCoreSelCPU0: Byte = $0; // AMD Athlon 64 Processors
  AMDCoreSelCPU1: Byte = $4;
  AMDMiscControlDeviceId: Word = MISCELLANEOUS_CONTROL_DEVICE_ID;
  AMDPackageTempProc: TPackageTempProc;
  AMDCoreTempProc: TCoreTempProc;


function CPUID: TCpuInfo;
  function HasCPUIDInstruction: Boolean;
  const
    ID_FLAG = $200000;
  {$IFNDEF DELPHI64_TEMPORARY}
  begin
  {$ENDIF ~DELPHI64_TEMPORARY}
    asm
      {$IFDEF CPU32}
      PUSHFD
      POP     EAX
      MOV     ECX, EAX
      XOR     EAX, ID_FLAG
      AND     ECX, ID_FLAG
      PUSH    EAX
      POPFD
      PUSHFD
      POP     EAX
      AND     EAX, ID_FLAG
      XOR     EAX, ECX
      SETNZ   Result
      {$ENDIF CPU32}
      {$IFDEF CPU64}
      {$IFDEF FPC}
        {$DEFINE DELPHI64_TEMPORARY}
      {$ENDIF FPC}
      {$IFDEF DELPHI64_TEMPORARY}
      PUSHFQ
      {$ELSE ~DELPHI64_TEMPORARY}
      PUSHFD
      {$ENDIF ~DELPHI64_TEMPORARY}
      POP     RAX
      MOV     RCX, RAX
      XOR     RAX, ID_FLAG
      AND     RCX, ID_FLAG
      PUSH    RAX
      {$IFDEF DELPHI64_TEMPORARY}
      POPFQ
      {$ELSE ~DELPHI64_TEMPORARY}
      POPFD
      {$ENDIF ~DELPHI64_TEMPORARY}
      {$IFDEF DELPHI64_TEMPORARY}
      PUSHFQ
      {$ELSE ~DELPHI64_TEMPORARY}
      PUSHFD
      {$ENDIF ~DELPHI64_TEMPORARY}
      POP     RAX
      AND     RAX, ID_FLAG
      XOR     RAX, RCX
      SETNZ   Result
      {$IFDEF FPC}
        {$UNDEF DELPHI64_TEMPORARY}
      {$ENDIF FPC}
      {$ENDIF CPU64}
    end;
  {$IFNDEF DELPHI64_TEMPORARY}
  end;
  {$ENDIF ~DELPHI64_TEMPORARY}

  procedure CallCPUID(ValueEAX, ValueECX: Cardinal; out ReturnedEAX, ReturnedEBX, ReturnedECX, ReturnedEDX);
  {$IFNDEF DELPHI64_TEMPORARY}
  begin
  {$ENDIF ~DELPHI64_TEMPORARY}
    asm
      {$IFDEF CPU32}
      // save context
      PUSH    EDI
      PUSH    EBX
      // init parameters
      MOV     EAX, ValueEAX
      MOV     ECX, ValueECX
      // CPUID
      DB      0FH
      DB      0A2H
      // store results
      MOV     EDI, ReturnedEAX
      MOV     Cardinal PTR [EDI], EAX
      MOV     EAX, ReturnedEBX
      MOV     EDI, ReturnedECX
      MOV     Cardinal PTR [EAX], EBX
      MOV     Cardinal PTR [EDI], ECX
      MOV     EAX, ReturnedEDX
      MOV     Cardinal PTR [EAX], EDX
      // restore context
      POP  EBX
      POP  EDI
      {$ENDIF CPU32}
      {$IFDEF CPU64}
      // save context
      PUSH    RBX
      // init parameters
      MOV     EAX, ValueEAX
      MOV     ECX, ValueECX
      // CPUID
      CPUID
      // store results
      MOV     R8, ReturnedEAX
      MOV     R9, ReturnedEBX
      MOV     R10, ReturnedECX
      MOV     R11, ReturnedEDX
      MOV     Cardinal PTR [R8], EAX
      MOV     Cardinal PTR [R9], EBX
      MOV     Cardinal PTR [R10], ECX
      MOV     Cardinal PTR [R11], EDX
      // restore context
      POP     RBX
      {$ENDIF CPU64}
    end;
  {$IFNDEF DELPHI64_TEMPORARY}
  end;
  {$ENDIF ~DELPHI64_TEMPORARY}

  procedure ProcessStandard(var CPUInfo: TCpuInfo; HiVal: Cardinal);
  var
    VersionInfo, AdditionalInfo, ExFeatures: Cardinal;
  begin
    if HiVal >= 1 then
    begin
      CallCPUID(1, 0, VersionInfo, AdditionalInfo, ExFeatures, CPUInfo.Features);

      CPUInfo.PType := (VersionInfo and $00003000) shr 12;
      CPUInfo.BaseFamily := (VersionInfo and $00000F00) shr 8;
      CPUInfo.BaseModel := (VersionInfo and $000000F0) shr 4;
      CPUInfo.Stepping := (VersionInfo and $0000000F);
      CPUInfo.ExtendedModel := (VersionInfo and $000F0000) shr 16;
      CPUInfo.ExtendedFamily := (VersionInfo and $0FF00000) shr 20;
        CPUInfo.Family := CPUInfo.BaseFamily;
        CPUInfo.Model := CPUInfo.BaseModel;

      if CPUInfo.CpuType = CPU_TYPE_INTEL then
      begin
        CPUInfo.IntelSpecific.ExFeatures := ExFeatures;
        CPUInfo.IntelSpecific.BrandID := AdditionalInfo and $000000FF;
        CPUInfo.IntelSpecific.APICID := (AdditionalInfo and $FF000000) shr 24;
        CPUInfo.HyperThreadingTechnology := (CPUInfo.Features and INTEL_HTT) <> 0;
        if CPUInfo.HyperThreadingTechnology then
        begin
          CPUInfo.LogicalCore := (AdditionalInfo and $00FF0000) shr 16;
          if CPUInfo.LogicalCore = 0 then
            CPUInfo.LogicalCore := 1;
        end;
      end;
    end;
  end;

  procedure ProcessIntel(var CPUInfo: TCpuInfo; HiVal: Cardinal);
  var
    ExHiVal, Unused, CoreInfo: Cardinal;
  begin
    CPUInfo.CpuType := CPU_TYPE_INTEL;
    CPUInfo.Manufacturer := ManufacturerINTEL;

    ProcessStandard(CPUInfo, HiVal);
    
    // Calc Family and Model with ExtendedFamily/ExtendedModel:
    // The Extended Family ID needs to be examined only when the Family ID is 0FH.
    //if CPUInfo.BaseFamily = $0F then
    //  CPUInfo.Family := CPUInfo.BaseFamily + CPUInfo.ExtendedFamily;
    
    // The Extended Model ID needs to be examined only when the Family ID is 06H or 0FH.
    //if CPUInfo.BaseFamily in [$06, $0F] then
    //  CPUInfo.Model := (CPUInfo.ExtendedModel shl 4) or CPUInfo.BaseModel;
    
    if HiVal >= 4 then
    begin
      CallCPUID(4, 0, CoreInfo, Unused, Unused, Unused);
      CPUInfo.PhysicalCore := ((CoreInfo and $FC000000) shr 26) + 1;
    end;
    
    // check Intel extended
    CallCPUID($80000000, 0, ExHiVal, Unused, Unused, Unused);
    if ExHiVal >= $80000001 then
    begin
      CPUInfo.HasExtendedInfo := True;
      CallCPUID($80000001, 0, Unused, Unused, CPUInfo.IntelSpecific.Ex64Features2,
        CPUInfo.IntelSpecific.Ex64Features);
    end;
    
    if ExHiVal >= $80000002 then
      CallCPUID($80000002, 0, CPUInfo.CpuName[0], CPUInfo.CpuName[4], CPUInfo.CpuName[8], CPUInfo.CpuName[12]);
    if ExHiVal >= $80000003 then
      CallCPUID($80000003, 0, CPUInfo.CpuName[16], CPUInfo.CpuName[20], CPUInfo.CpuName[24], CPUInfo.CpuName[28]);
    if ExHiVal >= $80000004 then
      CallCPUID($80000004, 0, CPUInfo.CpuName[32], CPUInfo.CpuName[36], CPUInfo.CpuName[40], CPUInfo.CpuName[44]);
    
    CPUInfo.HardwareHyperThreadingTechnology := CPUInfo.LogicalCore <> CPUInfo.PhysicalCore;
  end;

  procedure ProcessAMD(var CPUInfo: TCpuInfo; HiVal: Cardinal);
  var
    ExHiVal, Unused, VersionInfo, AdditionalInfo: Cardinal;
  begin
    CPUInfo.CpuType := CPU_TYPE_AMD;
    CPUInfo.Manufacturer := ManufacturerAMD;

    // check AMD extended
    if HiVal >= 1 then
    begin
      CallCPUID(1, 0, VersionInfo, AdditionalInfo, CPUInfo.AMDSpecific.Features2, CPUInfo.Features);

      CPUInfo.AMDSpecific.BrandID := AdditionalInfo and $000000FF;
      CPUInfo.AMDSpecific.APICID := (AdditionalInfo and $FF000000) shr 24;
      CPUInfo.HyperThreadingTechnology := (CPUInfo.Features and AMD_HTT) <> 0;
      if CPUInfo.HyperThreadingTechnology then
      begin
        CPUInfo.LogicalCore := (AdditionalInfo and $00FF0000) shr 16;
        if CPUInfo.LogicalCore = 0 then
          CPUInfo.LogicalCore := 1;
      end;
    end;

    CallCPUID($80000000, 0, ExHiVal, Unused, Unused, Unused);
    if ExHiVal <> 0 then
    begin
      // AMD only
      CPUInfo.HasExtendedInfo := True;

      if ExHiVal >= $80000001 then
      begin
        CallCPUID($80000001, 0, VersionInfo, AdditionalInfo, CPUInfo.AMDSpecific.ExFeatures2, CPUInfo.AMDSpecific.ExFeatures);
        CPUInfo.BaseFamily := (VersionInfo and $00000F00) shr 8;
        CPUInfo.BaseModel := (VersionInfo and $000000F0) shr 4;
        CPUInfo.Stepping := (VersionInfo and $0000000F);
        CPUInfo.ExtendedModel := (VersionInfo and $000F0000) shr 16;
        CPUInfo.ExtendedFamily := (VersionInfo and $0FF00000) shr 20;
          // Family and Model actually BaseFamily and BaseModel,
          // so calc with ExtendedFamily/ExtendedModel:
          CPUInfo.Family := CPUInfo.BaseFamily + CPUInfo.ExtendedFamily;
          CPUInfo.Model := (CPUInfo.ExtendedModel shl 4) or CPUInfo.BaseModel;
          
        CPUInfo.AMDSpecific.ExBrandID := AdditionalInfo and $0000FFFF;
      end;
      
      if ExHiVal >= $80000002 then
        CallCPUID($80000002, 0, CPUInfo.CpuName[0], CPUInfo.CpuName[4], CPUInfo.CpuName[8], CPUInfo.CpuName[12]);
      if ExHiVal >= $80000003 then
        CallCPUID($80000003, 0, CPUInfo.CpuName[16], CPUInfo.CpuName[20], CPUInfo.CpuName[24], CPUInfo.CpuName[28]);
      if ExHiVal >= $80000004 then
        CallCPUID($80000004, 0, CPUInfo.CpuName[32], CPUInfo.CpuName[36], CPUInfo.CpuName[40], CPUInfo.CpuName[44]);
      
      if ExHiVal >= $80000008 then
      begin
        CallCPUID($80000008, 0, Unused, VersionInfo, AdditionalInfo, Unused);
        CPUInfo.PhysicalCore := (AdditionalInfo and $000000FF) + 1;
      end;
    end
    else
    begin
      ProcessStandard(CPUInfo, HiVal);
    end;

    CPUInfo.HardwareHyperThreadingTechnology := CPUInfo.LogicalCore <> CPUInfo.PhysicalCore;
  end;


var
  HiVal: Cardinal;
begin
  FillChar(Result, SizeOf(Result), 0);
  Result.LogicalCore := 1;
  Result.PhysicalCore := 1;
  
  if HasCPUIDInstruction then
  begin
    Result.HasInstruction := True;
    CallCPUID(0, 0, HiVal, Result.VendorIDString[0], Result.VendorIDString[8],
      Result.VendorIDString[4]);
    if Result.VendorIDString = VendorIDIntel then
      ProcessIntel(Result, HiVal)
    else if Result.VendorIDString = VendorIDAMD then
      ProcessAMD(Result, HiVal)
    else
      ProcessStandard(Result, HiVal);
  end
  else
    Result.Family := 4;

  if Result.CpuType = 0 then
  begin
    Result.Manufacturer := 'Unknown';
    Result.CpuName := 'Unknown';
  end;
end;



function AMDGetPciAddress(Func: Byte; DeviceId: Word; processorIndex: Byte): DWord;
//var
//  Ven: DWord;
begin
  Result := PciBusDevFunc(PCI_BUS, (PCI_BASE_DEVICE + processorIndex), Func);
  
  //     
  
  // verify that we have the correct bus, device and function
  //Ven := ReadPciConfigDword(Result, DEVICE_VENDOR_ID_REGISTER);
  //if (Ven <> ((DeviceId shl 16) or AMD_VENDOR_ID)) then
  //  Result := InvalidPciAddress;
end;

function ReadAMD_0F_CoreTemp: TCoreTemps;
var
  I: Integer;
  T: Integer;
begin
  SetLength(Result, CpuInfo.PhysicalCore);
  //FillChar(Result[0], Length(Result)*SizeOf(Integer), 0);
  
  for I := 0 to Length(Result) - 1 do
  begin
    T := 0;
    WritePciConfigDwordEx(AMDPciAddress, THERMTRIP_STATUS_REGISTER, I*4); //  
    if ReadPciConfigDwordEx(AMDPciAddress, THERMTRIP_STATUS_REGISTER, @T) then
    begin
      T := (T shr 16) and $FF + AMDtOffset;
    end;
    
    Result[I] := T;
  end;
end;

function ReadAMD_10_CoreTemp: TCoreTemps;
var
  I: Integer;
  T: Integer;
  //Offs, Mask: Integer;
begin
  SetLength(Result, CpuInfo.PhysicalCore);
  //FillChar(Result[0], Length(Result)*SizeOf(Integer), 0);

  //    
  // TODO:       PrepareAMDParams()
  // ======================== //
  {Mask := $7FF;
  Offs := -49; // AMDtOffset  (Inc(AMDtOffset, 21)) ?
  
  with CPUInfo do
  begin
    if Family = $15 then
    begin
      if (Model and $F0) = 0 then Mask := $7FC;
    end
    else
      if Family <> $16 then Offs := 0;
  end;}
  // ======================== //
  
  for I := 0 to Length(Result) - 1 do
  begin
    T := 0;
    WritePciConfigDwordEx(AMDPciAddress, REPORTED_TEMPERATURE_CONTROL_REGISTER, I*4); //   (?)
    if ReadPciConfigDwordEx(AMDPciAddress, REPORTED_TEMPERATURE_CONTROL_REGISTER, @T) then
    begin
      //T := (((T shr 21) and Mask) div 8) + Offs;
      T := (((T shr 21) and $7FF) div 8);
    end;
    
    Result[I] := T;
  end;
end;

function ReadAMD_0F_PackageTemp: Integer;
var
  I: Integer;
  CT: TCoreTemps;
begin
  Result := 0;
  
  //    -     .
  //   .
  CT := ReadAMD_0F_CoreTemp;
  for I := 0 to Length(CT) - 1 do
    if CT[I] > Result then Result := CT[I];
end;

function ReadAMD_10_PackageTemp: Integer;
begin
  Result := 0;
  //    -    (?)
  if ReadPciConfigDwordEx(AMDPciAddress, REPORTED_TEMPERATURE_CONTROL_REGISTER, @Result)
  then
  begin
    with CPUInfo do
    begin
      if (Family = $15) and ((Result and $30000) <> 0) then
      begin
        if (Model and $F0) = 0 then
          Result := (((Result shr 21) and $7FC) div 8) - 49
        else
          Result := (((Result shr 21) and $7FF) div 8) - 49;
      end
      else
        if (Family = $16) and (((Result and $30000) <> 0) or ((Result and $80000) <> 0))
        then
          Result := (((Result shr 21) and $7FF) div 8) - 49
        else
          Result := (((Result shr 21) and $7FF) div 8);
    end;
  end;
end;



function GetTjMaxFromMSR(): Integer;
var
  eax: DWord;
begin
  if WinRing0.RdmsrTx(MSR_TEMPERATURE_TARGET, @eax, nil, 1)
  then
  begin
    Result := (eax shr 16) and $FF;
  end
  else
    Result := 100;
end;

function ReadIntelCoreTemp: TCoreTemps;
var
  I, Step: Integer;
  hThread: THandle;
  Mask: DWord;
  eax: DWord;
begin
  SetLength(Result, CpuInfo.PhysicalCore);
  FillChar(Result[0], Length(Result)*SizeOf(Integer), 0);
  
  //      CPU -   
  if IntelTJMax = 0 then IntelTJMax := GetTjMaxFromMSR();
  
  hThread := GetCurrentThread();
  Mask := SetThreadAffinityMask(hThread, 1);
  
  //   HT 1   1    2  -
  //     ,   AffinityMask    
  Step := CpuInfo.LogicalCore div CpuInfo.PhysicalCore;
  
  for I := 0 to Length(Result) - 1 do
  begin
    SetThreadAffinityMask(hThread, (1 shl I*Step));
    if WinRing0.Rdmsr(IA32_THERM_STATUS, @eax) then
    begin    
      Result[I] := IntelTJMax - Integer(1 * ((eax and $007F0000) shr 16));
    end
    else
    begin
      //SetLength(Result, 0);
      Break;
    end;
  end;
  SetThreadAffinityMask(hThread, Mask);
end;


function ReadIntelPackageTemp: Integer;
var
  eax: DWord;
  I: Integer;
  CT: TCoreTemps;
begin
  Result := 0;
  CT := nil;
  
  //      CPU -   
  if IntelTJMax = 0 then IntelTJMax := GetTjMaxFromMSR();
  
  if WinRing0.RdmsrTx(IA32_PACKAGE_THERM_STATUS, @eax, nil, 1)
  then
  begin
    // get the dist from tjMax from bits 22:16
    I := ((eax and $007F0000) shr 16); // = deltaT
    
    Result := IntelTJMax - IntelTSlope * I;
  end
  else
  begin
    //     .
    //   .
    CT := ReadIntelCoreTemp;
    for I := 0 to Length(CT) - 1 do
      if CT[I] > Result then Result := CT[I];
  end;
end;


  
procedure PrepareIntelParams;
begin
  with CPUInfo do
  begin
    case BaseFamily of
      $06:
        case BaseModel of
          $0F: // Intel Core 2 (65nm)
            case Stepping of
              $06: // B2
                {case PhysicalCore of
                  2:
                    IntelTJMax := (80 + 10);
                  4:
                    IntelTJMax := (90 + 10);
                  else
                    IntelTJMax := (85 + 10);
                end;}
                IntelTJMax := (80 + 10);
              $0B: // G0
                IntelTJMax := (90 + 10); 
              $0D: // M0
                IntelTJMax := (85 + 10);
              else
                IntelTJMax := (85 + 10);
            end;
          
          $17: // Intel Core 2 (45nm)
            IntelTJMax := 100;
          $1C: // Intel Atom (45nm)
            case Stepping of
              $02: // C0
                IntelTJMax := 90;
              $0A: // A0, B0
                IntelTJMax := 100;
              else
                IntelTJMax := 90;
            end;
          //  Open Hardware Monitor
          $1A, // Intel Core i7 LGA1366 (45nm)
          $1E, // Intel Core i5, i7 LGA1156 (45nm)
          $1F, // Intel Core i5, i7 
          $25, // Intel Core i3, i5, i7 LGA1156 (32nm)
          $2A, // Intel Core i5, i7 2xxx LGA1155 (32nm)
          $2C, // Intel Core i7 LGA1366 (32nm) 6 Core
          $2D, // Next Generation Intel Xeon, i7 3xxx LGA2011 (32nm)
          $2E, // Intel Xeon Processor 7500 series (45nm)
          $2F, // Intel Xeon Processor (32nm)
          $3A, // Intel Core i5, i7 3xxx LGA1155 (22nm)
          $3C, // Intel Core i5, i7 4xxx LGA1150 (22nm)
          $3E, // Intel Core i7 4xxx LGA2011 (22nm)
          $45,
          $46:
            IntelTJMax := 0; //     GetTjMaxFromMSR();
          
          //   Intel
          { $1A,$1E,$1F,$2A,$2D,$2E,}$37,{ $3E,}
          $4A,$4D,$4F,$56,$57,$5A,$5C,$5D,$7A:
            //IntelTJMax := 0;
          else
            IntelTJMax := 100;
        end;
      
      $0F:
        (*case Model of
          {$00, // Pentium 4 (180nm)
          $01, // Pentium 4 (130nm)
          $02, // Pentium 4 (130nm)
          $03, // Pentium 4, Celeron D (90nm)
          $04, // Pentium 4, Pentium D, Celeron D (90nm)
          $06: // Pentium 4, Pentium D, Celeron D (65nm)}
          $00..$06:
            IntelTJMax := 100;
          else
            IntelTJMax := 100;
        end;*)
        IntelTJMax := 100;
        
      else // Family of
        IntelTJMax := 100;
    end;
  end;
end;

procedure PrepareAMDParams;
begin
  with CPUInfo do
  begin
    // AM2+ 65nm +21 offset
    // - :  AMD Athlon(tm) 64 X2 Dual Core Processor 4800+
    // (Family: 0F, Model: 0B, Stepping: 2, ExtendedFamily: 00, ExtendedModel: 06)
    //   Inc(AMDtOffset, 21),    
    {if (Model >= $69) and 
       not (Model in [$c1, $6c, $7c])
    then}
      Inc(AMDtOffset, 21);
      
    case Family of
      $10: AMDMiscControlDeviceId := FAMILY_10H_MISCELLANEOUS_CONTROL_DEVICE_ID;
      $11: AMDMiscControlDeviceId := FAMILY_11H_MISCELLANEOUS_CONTROL_DEVICE_ID;
      $12: AMDMiscControlDeviceId := FAMILY_12H_MISCELLANEOUS_CONTROL_DEVICE_ID;
      $14: AMDMiscControlDeviceId := FAMILY_14H_MISCELLANEOUS_CONTROL_DEVICE_ID;
      $15: 
        case Model of
          $00: AMDMiscControlDeviceId := FAMILY_15H_MODEL_00_MISC_CONTROL_DEVICE_ID;
          $10: AMDMiscControlDeviceId := FAMILY_15H_MODEL_10_MISC_CONTROL_DEVICE_ID;
        end;
      $16:
        case Model of
          $00: AMDMiscControlDeviceId := FAMILY_16H_MODEL_00_MISC_CONTROL_DEVICE_ID;
        end;
    end;
      
    if Family in [$10..$16] then
    begin
      AMDPackageTempProc := ReadAMD_10_PackageTemp;
      AMDCoreTempProc := ReadAMD_10_CoreTemp;
    end
    else
    begin
      AMDPackageTempProc := ReadAMD_0F_PackageTemp;
      AMDCoreTempProc := ReadAMD_0F_CoreTemp;
    end;
    
    // AMD NPT Family 0Fh Revision F, G have the core selection swapped
    {if (Model >= 40) then
    begin
      AMDCoreSelCPU0 := $4;
      AMDCoreSelCPU1 := $0;
    end;}
  end;
  
  AMDPciAddress := AMDGetPciAddress(
    MISCELLANEOUS_CONTROL_FUNCTION, AMDMiscControlDeviceId, 
    AMDCoreSelCPU0
  );
end;

procedure InitTemperatureParams;
begin
  case CPUInfo.CpuType of
    CPU_TYPE_INTEL: PrepareIntelParams;
    CPU_TYPE_AMD: PrepareAMDParams;
  end;
end;


function GetCPUPackageTemp: Integer;
begin
  case CPUInfo.CpuType of
    CPU_TYPE_INTEL: Result := ReadIntelPackageTemp();
    CPU_TYPE_AMD: Result := AMDPackageTempProc();
    else
      Result := 0;
  end;
end;

function GetCPUCoreTemp: TCoreTemps;
begin
  case CPUInfo.CpuType of
    CPU_TYPE_INTEL: Result := ReadIntelCoreTemp;
    CPU_TYPE_AMD: Result := AMDCoreTempProc();
    else
    begin
      SetLength(Result, CpuInfo.PhysicalCore);
      FillChar(Result[0], Length(Result)*SizeOf(Integer), 0);
    end;
  end;
end;


constructor ThiCPUTemperature.Create;
begin
  if not WinRing0.OpenDriver() then
  begin
    FDrvStatus := GetLastError();
    //if FDrvStatus = 0 then FDrvStatus := -1; //   
  end;
  inherited;
end;


destructor ThiCPUTemperature.Destroy;
begin
  WinRing0.CloseDriver();
  inherited;
end;


procedure ThiCPUTemperature._work_doGetPackageTemp(var _Data: TData; Index: Word);
begin
  if FDrvStatus <> 0 then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorDriver, FDrvStatus);
    Exit;
  end;
  
  FTemp := GetCPUPackageTemp();
  _hi_CreateEvent(_Data, @_event_onGetPackageTemp, FTemp);
end;

procedure ThiCPUTemperature._work_doGetCoreTemp(var _Data: TData; Index: Word);
var
  CT: TCoreTemps;
  I: Integer;
  MTData: array of TData;
begin
  if FDrvStatus <> 0 then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorDriver, FDrvStatus);
    Exit;
  end;
  
  CT := GetCPUCoreTemp;
  if Length(CT) = 0 then Exit;
  
  SetLength(MTData, Length(CT));
  
  //   TData
  for I := 0 to Length(CT) - 1 do
  begin
    dtInteger(MTData[I], CT[I]);
  end;
  
  //  "MT-" - MTData[I].ldata     MTData[I+1]
  for I := 1 to Length(CT) - 1 do
  begin
    MTData[I-1].ldata := @MTData[I];
  end;
  
  _hi_OnEvent(_event_onGetCoreTemp, MTData[0]);
end;

procedure ThiCPUTemperature._var_CPUTemp(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, FTemp);
end;

procedure ThiCPUTemperature._var_VendorID(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, CpuInfo.CpuType);
end;

procedure ThiCPUTemperature._var_CPUName(var _Data: TData; Index: Word);
begin
  dtString(_Data, CpuInfo.CpuName);
end;

procedure ThiCPUTemperature._var_Family(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, CpuInfo.Family);
end;

procedure ThiCPUTemperature._var_Model(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, CpuInfo.Model);
end;

procedure ThiCPUTemperature._var_Stepping(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, CpuInfo.Stepping);
end;

procedure ThiCPUTemperature._var_PhysicalCores(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, CpuInfo.PhysicalCore);
end;

procedure ThiCPUTemperature._var_LogicalCores(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, CpuInfo.LogicalCore);
end;

initialization
  CpuInfo := CPUID;
  InitTemperatureParams();


end.